##  Copyright (c) 2022 Rational Commerce Ltd.

'''
Python bindings for the Actian Ingres OpenAPI
'''


from functools import wraps
import ctypes as C
from loguru import logger
import os
from sys import platform


from .IIAPI_CONSTANTS import *
from .IIAPI_PARM import *


_name = __name__


##  pick up environment settings
IIAPI_DEV_MODE = ('IIAPI_DEV_MODE' in os.environ
    and os.environ['IIAPI_DEV_MODE'] == 'ON')
IIAPI_DEBUG_ONERROR = ('IIAPI_DEBUG_ONERROR' in os.environ 
    and os.environ['IIAPI_DEBUG_ONERROR'] == 'ON')


##  IIAPI_DEV_MODE=ON puts pyngres in development mode
if IIAPI_DEV_MODE:
    import importlib.metadata
    _version = importlib.metadata.version(_name)
    logger.warning(f'using {_name} {_version} in development mode')
    logger.info(f'to disable logging messages: logger.disable(\'{_name}\')')
    logger.info(f'to enable logging messages: logger.enable(\'{_name}\')')
else:
    logger.disable(_name)


##  load the Ingres OpenAPI using ctypes
if platform == 'linux':
    ##  expect to find libiiapi.1.so in $II_SYSTEM/ingres/lib
    try:
        II_SYSTEM = os.environ['II_SYSTEM']
    except KeyError:
        print('II_SYSTEM is not set in the environment')
        quit()
    except Exception as exception:
        print(exception)
        quit()
    api_pathname = f'{II_SYSTEM}/ingres/lib/libiiapi.1.so'
elif platform == 'win32':
    ##  expect to find iilibapi.dll in ..\ingres\bin
    ##  II_SYSTEM isn't necessarily set in the Windows environment; search
    ##  PATH for ingres\bin
    PATH = os.environ['PATH']
    paths = PATH.split(';')
    folders = [
        folder for folder in paths
        if folder.endswith('ingres\\bin') and '.' not in folder]
    nfolders = len(folders)
    if nfolders == 0:
        print('ingres\\bin is not defined in PATH')
        quit()
    ##  use the first match
    path = folders[0]
    api_pathname = f'{path}\\iilibapi.dll'
elif platform == 'darwin':
    ##  expect to find libq.1.dylib and libiiapi.1.dylib in 
    ##  $II_SYSTEM/ingres/lib; libq.1.dylib needs to be loaded before loading
    ##  libiiapai.1.dylib
    try:
        II_SYSTEM = os.environ['II_SYSTEM']
    except KeyError:
        print('II_SYSTEM is not set in the environment')
        quit()
    except Exception as exception:
        print(exception)
        quit()
    api_pathname = f'{II_SYSTEM}/ingres/lib/libiiapi.1.dylib'
    ##  load libq so it's ready when we load the OpenAPI
    libq_pathname = f'{II_SYSTEM}/ingres/lib/libq.1.dylib'
    logger.debug(f'attempting to load {libq_pathname}')
    libq = C.CDLL(libq_pathname, mode=C.RTLD_GLOBAL)
    logger.success(f'loaded {libq_pathname}')
else:
    print(f'No python binding is available for Ingres OpenAPI on {platform}.')
    print('(If the API exists, a binding is easy to add. Give us a call.)')
    quit()
##  don't use try/except here; just let it explode if it wants to
logger.debug(f'attempting to load {api_pathname}')
iiapi = C.CDLL(api_pathname)
logger.success(f'loaded {api_pathname}')


##  decorator to simplify the binding of functions
def bind_function(lib, funcname, restype, argtypes):
    '''return a python wrapper for a C function'''
    func = lib.__getattr__(funcname)
    func.restype = restype
    func.argtypes = argtypes
    return func


##  bind the Ingres OpenAPI functions
IIapi_abort = bind_function(
    iiapi, 'IIapi_abort', None, [C.POINTER(IIAPI_ABORTPARM)]
)
IIapi_autocommit = bind_function(
    iiapi, 'IIapi_autocommit', None, [C.POINTER(IIAPI_AUTOPARM)]
)
IIapi_batch = bind_function(
    iiapi, 'IIapi_batch', None, [C.POINTER(IIAPI_BATCHPARM)]
)
IIapi_cancel = bind_function(
    iiapi, 'IIapi_cancel', None, [C.POINTER(IIAPI_CANCELPARM)]
)
IIapi_catchEvent = bind_function(
    iiapi, 'IIapi_catchEvent', None, [C.POINTER(IIAPI_CATCHEVENTPARM)]
)
IIapi_close = bind_function(
    iiapi, 'IIapi_close', None, [C.POINTER(IIAPI_CLOSEPARM)]
)
IIapi_commit = bind_function(
    iiapi, 'IIapi_commit', None, [C.POINTER(IIAPI_COMMITPARM)]
)
IIapi_connect = bind_function(
    iiapi, 'IIapi_connect', None, [C.POINTER(IIAPI_CONNPARM)]
)
IIapi_convertData = bind_function(
    iiapi, 'IIapi_convertData', None, [C.POINTER(IIAPI_CONVERTPARM)]
)
IIapi_disconnect = bind_function(
    iiapi, 'IIapi_disconnect', None, [C.POINTER(IIAPI_DISCONNPARM)]
)
IIapi_formatData = bind_function(
    iiapi, 'IIapi_formatData', None, [C.POINTER(IIAPI_FORMATPARM)]
)
IIapi_getColumnInfo = bind_function(
    iiapi, 'IIapi_getColumnInfo', None, [C.POINTER(IIAPI_GETCOLINFOPARM)]
)
IIapi_getColumns = bind_function(
    iiapi, 'IIapi_getColumns', None, [C.POINTER(IIAPI_GETCOLPARM)]
)
IIapi_getCopyMap = bind_function(
    iiapi, 'IIapi_getCopyMap', None, [C.POINTER(IIAPI_GETCOPYMAPPARM)]
)
IIapi_getDescriptor = bind_function(
    iiapi, 'IIapi_getDescriptor', None, [C.POINTER(IIAPI_GETDESCRPARM)]
)
IIapi_getErrorInfo = bind_function(
    iiapi, 'IIapi_getErrorInfo', None, [C.POINTER(IIAPI_GETEINFOPARM)]
)
IIapi_getEvent = bind_function(
    iiapi, 'IIapi_getEvent', None, [C.POINTER(IIAPI_GETEVENTPARM)]
)
IIapi_getQueryInfo = bind_function(
    iiapi, 'IIapi_getQueryInfo', None, [C.POINTER(IIAPI_GETQINFOPARM)]
)
IIapi_initialize = bind_function(
    iiapi, 'IIapi_initialize', None, [C.POINTER(IIAPI_INITPARM)]
)
IIapi_modifyConnect = bind_function(
    iiapi, 'IIapi_modifyConnect', None, [C.POINTER(IIAPI_MODCONNPARM)]
)
IIapi_position = bind_function(
    iiapi, 'IIapi_position', None, [C.POINTER(IIAPI_POSPARM)]
)
IIapi_prepareCommit = bind_function(
    iiapi, 'IIapi_prepareCommit', None, [C.POINTER(IIAPI_PREPCMTPARM)]
)
IIapi_putColumns = bind_function(
    iiapi, 'IIapi_putColumns', None, [C.POINTER(IIAPI_PUTCOLPARM)]
)
IIapi_putParms = bind_function(
    iiapi, 'IIapi_putParms', None, [C.POINTER(IIAPI_PUTPARMPARM)]
)
IIapi_query = bind_function(
    iiapi, 'IIapi_query', None, [C.POINTER(IIAPI_QUERYPARM)]
)
IIapi_registerXID = bind_function(
    iiapi, 'IIapi_registerXID', None, [C.POINTER(IIAPI_REGXIDPARM)]
)
IIapi_releaseEnv = bind_function(
    iiapi, 'IIapi_releaseEnv', None, [C.POINTER(IIAPI_RELENVPARM)]
)
IIapi_releaseXID = bind_function(
    iiapi, 'IIapi_releaseXID', None, [C.POINTER(IIAPI_RELXIDPARM)]
)
IIapi_rollback = bind_function(
    iiapi, 'IIapi_rollback', None, [C.POINTER(IIAPI_ROLLBACKPARM)]
)
IIapi_savePoint = bind_function(
    iiapi, 'IIapi_savePoint', None, [C.POINTER(IIAPI_SAVEPTPARM)]
)
IIapi_scroll = bind_function(
    iiapi, 'IIapi_scroll', None, [C.POINTER(IIAPI_SCROLLPARM)]
)
IIapi_setConnectParam = bind_function(
    iiapi, 'IIapi_setConnectParam', None, [C.POINTER(IIAPI_SETCONPRMPARM)]
)
IIapi_setDescriptor = bind_function(
    iiapi, 'IIapi_setDescriptor', None, [C.POINTER(IIAPI_SETDESCRPARM)]
)
IIapi_setEnvParam = bind_function(
    iiapi, 'IIapi_setEnvParam', None, [C.POINTER(IIAPI_SETENVPRMPARM)]
)
IIapi_terminate = bind_function(
    iiapi, 'IIapi_terminate', None, [C.POINTER(IIAPI_TERMPARM)]
)
IIapi_wait = bind_function(
    iiapi, 'IIapi_wait', None, [C.POINTER(IIAPI_WAITPARM)]
)
IIapi_xaCommit = bind_function(
    iiapi, 'IIapi_xaCommit', None, [C.POINTER(IIAPI_XACOMMITPARM)]
)
IIapi_xaStart = bind_function(
    iiapi, 'IIapi_xaStart', None, [C.POINTER(IIAPI_XASTARTPARM)]
)
IIapi_xaEnd = bind_function(
    iiapi, 'IIapi_xaEnd', None, [C.POINTER(IIAPI_XAENDPARM)]
)
IIapi_xaPrepare = bind_function(
    iiapi, 'IIapi_xaPrepare', None, [C.POINTER(IIAPI_XAPREPPARM)]
)
IIapi_xaRollback = bind_function(
    iiapi, 'IIapi_xaRollback', None, [C.POINTER(IIAPI_XAROLLPARM)]
)


##  OpenAPI callback assistance functions

IIapi_callback = C.CFUNCTYPE( None, C.POINTER(C.py_object), IIAPI_GENPARM )


def IIapi_getCallbackPtr(callback):
    '''return c_void_p pointer to an OpenAPI-compliant Python callback'''
    ptr =  C.cast(callback,C.c_void_p)
    return ptr


def IIapi_getClosurePtr(object):
    '''return c_void_p pointer to Python object to use as callback closure'''
    closure = C.pointer(C.py_object(object))
    ptr = C.cast(closure,C.c_void_p)
    return ptr


def IIapi_getClosure(ptr):
    '''return the Python object pointed to by C.c_void_p pointer'''
    object = C.cast(ptr,C.POINTER(C.py_object))
    closure = object.contents.value
    return closure


##  trace the OpenAPI calls when IIAPI_DEV_MODE envar is set

def traced(iiapi_func):
    '''decorator to trace (OpenAPI) calls'''

    @wraps(iiapi_func)
    def api_trace(*args, **kwargs):
        logger.opt(depth=1).trace(f'calling {iiapi_func.__name__}()...')
        result = iiapi_func(*args, **kwargs)
        logger.opt(depth=1).trace(f'... {iiapi_func.__name__} returned')
        return result
    return api_trace


if IIAPI_DEV_MODE:
    IIapi_abort = traced(IIapi_abort)
    IIapi_autocommit = traced(IIapi_autocommit)
    IIapi_batch = traced(IIapi_batch)
    IIapi_cancel = traced(IIapi_cancel)
    IIapi_catchEvent = traced(IIapi_catchEvent)
    IIapi_close = traced(IIapi_close)
    IIapi_commit = traced(IIapi_commit)
    IIapi_connect = traced(IIapi_connect)
    IIapi_convertData = traced(IIapi_convertData)
    IIapi_disconnect = traced(IIapi_disconnect)
    IIapi_formatData = traced(IIapi_formatData)
    IIapi_getColumnInfo = traced(IIapi_getColumnInfo)
    IIapi_getColumns = traced(IIapi_getColumns)
    IIapi_getCopyMap = traced(IIapi_getCopyMap)
    IIapi_getDescriptor = traced(IIapi_getDescriptor)
    IIapi_getErrorInfo = traced(IIapi_getErrorInfo)
    IIapi_getEvent = traced(IIapi_getEvent)
    IIapi_getQueryInfo = traced(IIapi_getQueryInfo)
    IIapi_initialize = traced(IIapi_initialize)
    IIapi_modifyConnect = traced(IIapi_modifyConnect)
    IIapi_position = traced(IIapi_position)
    IIapi_prepareCommit = traced(IIapi_prepareCommit)
    IIapi_putColumns = traced(IIapi_putColumns)
    IIapi_putParms = traced(IIapi_putParms)
    IIapi_query = traced(IIapi_query)
    IIapi_registerXID = traced(IIapi_registerXID)
    IIapi_releaseEnv = traced(IIapi_releaseEnv)
    IIapi_releaseXID = traced(IIapi_releaseXID)
    IIapi_rollback = traced(IIapi_rollback)
    IIapi_savePoint = traced(IIapi_savePoint)
    IIapi_scroll = traced(IIapi_scroll)
    IIapi_setConnectParam = traced(IIapi_setConnectParam)
    IIapi_setDescriptor = traced(IIapi_setDescriptor)
    IIapi_setEnvParam = traced(IIapi_setEnvParam)
    IIapi_terminate = traced(IIapi_terminate)
    IIapi_wait = traced(IIapi_wait)
    IIapi_xaCommit = traced(IIapi_xaCommit)
    IIapi_xaStart = traced(IIapi_xaStart)
    IIapi_xaEnd = traced(IIapi_xaEnd)
    IIapi_xaPrepare = traced(IIapi_xaPrepare)
    IIapi_xaRollback = traced(IIapi_xaRollback)


def debugged(function):
    '''decorator to debug on error'''
    ##  WARNING: using IIAPI_DEBUG_ONERROR calls all OpenAPI functions
    ##  synchronously, so pyngres.asyncio performance will be degraded
    def break_on_error(arg):
        ##  does arg contain an IIAPI_GENPARM? (i.e. function is asynchronous)?
        genParm = arg.genParm()
        if genParm:
            ##  function is asynchronous; status not available until completed
            function(arg)
            wtp = IIAPI_WAITPARM()
            wtp.wt_timeout = -1
            while not genParm.gp_completed:
                IIapi_wait(wtp)
            status = genParm.gp_status
        else:
            ##  function is synchronous; get status by suffix
            function(arg)
            status = arg.field_by_suffix('status')

        if status and status >= IIAPI_ST_ERROR:
            name = function.__name__
            logger.error(f'{name}() returned {status=}')
            print(arg)
            logger.info('(Pdb) is the debugger prompt')
            logger.info('to advance to the next statement press "n" then enter')
            logger.info('to continue/resume executing press "c" then enter')
            logger.info('for help press "h" then enter')
            breakpoint()
    return break_on_error
 

##  IIAPI_DEBUG_ONERROR=ON causes pyngres to invoke the pdb degugger
##  when an OpenAPI error is detected
if IIAPI_DEBUG_ONERROR:
    IIapi_abort = debugged(IIapi_abort)
    IIapi_autocommit = debugged(IIapi_autocommit)
    IIapi_batch = debugged(IIapi_batch)
    IIapi_cancel = debugged(IIapi_cancel)
    IIapi_catchEvent = debugged(IIapi_catchEvent)
    IIapi_close = debugged(IIapi_close)
    IIapi_commit = debugged(IIapi_commit)
    IIapi_connect = debugged(IIapi_connect)
    IIapi_convertData = debugged(IIapi_convertData)
    IIapi_disconnect = debugged(IIapi_disconnect)
    IIapi_formatData = debugged(IIapi_formatData)
    IIapi_getColumnInfo = debugged(IIapi_getColumnInfo)
    IIapi_getColumns = debugged(IIapi_getColumns)
    IIapi_getCopyMap = debugged(IIapi_getCopyMap)
    IIapi_getDescriptor = debugged(IIapi_getDescriptor)
    IIapi_getErrorInfo = debugged(IIapi_getErrorInfo)
    IIapi_getEvent = debugged(IIapi_getEvent)
    IIapi_getQueryInfo = debugged(IIapi_getQueryInfo)
    IIapi_initialize = debugged(IIapi_initialize)
    IIapi_modifyConnect = debugged(IIapi_modifyConnect)
    IIapi_position = debugged(IIapi_position)
    IIapi_prepareCommit = debugged(IIapi_prepareCommit)
    IIapi_putColumns = debugged(IIapi_putColumns)
    IIapi_putParms = debugged(IIapi_putParms)
    IIapi_query = debugged(IIapi_query)
    IIapi_registerXID = debugged(IIapi_registerXID)
    IIapi_releaseEnv = debugged(IIapi_releaseEnv)
    IIapi_releaseXID = debugged(IIapi_releaseXID)
    IIapi_rollback = debugged(IIapi_rollback)
    IIapi_savePoint = debugged(IIapi_savePoint)
    IIapi_scroll = debugged(IIapi_scroll)
    IIapi_setConnectParam = debugged(IIapi_setConnectParam)
    IIapi_setDescriptor = debugged(IIapi_setDescriptor)
    IIapi_setEnvParam = debugged(IIapi_setEnvParam)
    IIapi_terminate = debugged(IIapi_terminate)
    IIapi_wait = debugged(IIapi_wait)
    IIapi_xaCommit = debugged(IIapi_xaCommit)
    IIapi_xaStart = debugged(IIapi_xaStart)
    IIapi_xaEnd = debugged(IIapi_xaEnd)
    IIapi_xaPrepare = debugged(IIapi_xaPrepare)
    IIapi_xaRollback = debugged(IIapi_xaRollback)
    
    logger.warning('on-error debugging using pdb is enabled')



import ctypes as C
import os
import struct
from loguru import logger

from .IIAPI_CONSTANTS import *


##  the following are known from iiapidep.h
II_BOOL = C.c_int  ##  II_BOOL is not what you might guess...
II_CHAR = C.c_char
II_FLOAT4 = C.c_float
II_FLOAT8 = C.c_double
II_INT = C.c_int
II_INT1 = C.c_char
II_INT2 = C.c_short
II_INT4 = C.c_int
II_INT8 = C.c_longlong
II_LONG = C.c_int
II_PTR = C.c_void_p
II_STR = C.c_char_p  ##  use for e.g. II_CHAR II_FAR *string
II_UCHAR = C.c_ubyte
II_UINT1 = C.c_ubyte
II_UINT2 = C.c_ushort
II_UINT4 = C.c_uint
II_UINT8 = C.c_ulonglong
II_ULONG = C.c_uint

IIAPI_DT_ID = II_INT2
IIAPI_QUERYTYPE = II_ULONG
IIAPI_STATUS = II_ULONG


##  defining the IIAPI_DEV_MODE envar puts the API in development mode
if 'IIAPI_DEV_MODE' in os.environ:
    IIAPI_DEV_MODE = True
else:
    IIAPI_DEV_MODE = False


class IIAPI(C.Structure):
    '''structures for the Ingres OpenAPI'''

    if IIAPI_DEV_MODE:
        ##  this checking is only useful during development
        def __setattr__(self, name, value):
            '''prevent undeclared attributes from being set'''
            names = [name for name, type in self._fields_]
            if name not in names:
                struct_name = self.__class__.__name__
                diagnostic = '{struct_name} object has no attribute "{name}"'
                raise AttributeError(diagnostic)
            super().__setattr__(name, value)

    display_trim = ''


    ##  all the OpenAPI structure member names have a structure-specific
    ##  prefix (e.g. co_genParm, qy_genparm, etc). The following methods
    ##  facilitate access when the prefix is not known, using only the
    ##  suffix (e.g. genParm)

    def _field_name_prefix(self):
        '''return the field name prefix used in the structure'''

        prefix = None
        field = self._fields_[0]
        name = field[0]
        prefix = name[0:2]
        return prefix


    def field_by_suffix(self,suffix):
        '''get a structure field using only its name-suffix'''

        prefix = self._field_name_prefix()
        full_name = prefix + '_' + suffix
        try:
            field = getattr(self,full_name)
        except AttributeError:
            field = None
        return field


    def genParm(self):
        '''return the genParm field (if any)'''

        return self.field_by_suffix('genParm')


    def connHandle(self):
        '''return the connHandle field (if any)'''

        return self.field_by_suffix('connHandle')


    def tranHandle(self):
        '''return the tranHandle field (if any)'''

        return self.field_by_suffix('tranHandle')


    def stmtHandle(self):
        '''return the stmtHandle field (if any)'''

        return self.field_by_suffix('stmtHandle')


    def __repr__(self):
        struct_name = self.__class__.__name__
        repr = []
        repr.append('')
        repr.append(f'{struct_name}:')
        for name, type in self._fields_:
            value = getattr(self, name)
            if type == II_PTR:
                handle = f'{value:#018x}' if value else 'None'
                repr.append(f'{name}={handle}')
            else:
                repr.append(f'{name}={value}')
        separator = '\n' + self.display_trim
        return separator.join(repr)


class IIAPI_MEMBER(IIAPI):
    '''data structures used as members by the OpenAPI'''

    display_trim = '||  '


#class IIAPI_GENERIC(object):
#    '''mixin class to return *_genParm, *_connHandle, etc'''

class IIAPI_GENPARM(IIAPI_MEMBER):
    '''common parameter member for Ingres OpenAPI'''

    _fields_ = [
        ('gp_callback', II_PTR),
        ('gp_closure', II_PTR),
        ('gp_completed', II_BOOL),
        ('gp_status', IIAPI_STATUS),
        ('gp_errorHandle', II_PTR),
    ]


class IIAPI_DESCRIPTOR(IIAPI_MEMBER):
    '''The IIAPI_DESCRIPTOR data type describes OpenAPI data, including
    its type, length, precision, scale, and usage.'''

    _fields_ = [
        ('ds_dataType', IIAPI_DT_ID),
        ('ds_nullable', II_BOOL),
        ('ds_length', II_UINT2),
        ('ds_precision', II_INT2),
        ('ds_scale', II_INT2),
        ('ds_columnType', II_INT2),
        ('ds_columnName', II_STR),
    ]


class IIAPI_DATAVALUE(IIAPI_MEMBER):
    '''generic column data'''

    _fields_ = [
        ('dv_null', II_BOOL), 
        ('dv_length', II_UINT2), 
        ('dv_value', II_PTR)
    ]


class IIAPI_FDATADESCR(IIAPI_MEMBER):
    '''
    This data type describes the data in a copy file. It also describes
    how the file should be formatted.
    '''

    _fields_ = [
        ('fd_name', C.c_char_p),
        ('fd_type', C.c_short),
        ('fd_length', C.c_short),
        ('fd_prec', C.c_short),
        ('fd_column', C.c_int),
        ('fd_funcID', C.c_int),
        ('fd_cvLen', C.c_int),
        ('fd_cvPrec', C.c_int),
        ('fd_delimiter', C.c_bool),
        ('fd_delimLength', C.c_short),
        ('fd_delimValue', C.c_char_p),
        ('fd_nullable', C.c_bool),
        ('fd_nullInfo', C.c_bool),
        ('fd_nullDescr', IIAPI_DESCRIPTOR),
        ('fd_nullValue', IIAPI_DATAVALUE),
    ]


class IIAPI_COPYMAP(IIAPI_MEMBER):
    '''
    The IIAPI_COPYMAP data type provides information needed
    to execute the copy statement, including the copy file name,
    log file name, number of columns in a row, and a description
    of the data.
    '''

    _fields_ = [
        ('cp_copyFrom', II_BOOL),
        ('cp_flags', II_ULONG),
        ('cp_errorCount', II_LONG),
        ('cp_fileName', II_STR),
        ('cp_logName', II_STR),
        ('cp_dbmsCount', II_INT2),
        ('cp_dbmsDescr', C.POINTER(IIAPI_DESCRIPTOR)),
        ('cp_fileCount', II_INT2),
        ('cp_fileDescr', C.POINTER(IIAPI_FDATADESCR)),
    ]


class IIAPI_SVR_ERRINFO(IIAPI_MEMBER):
    '''additional server information associated with error messages'''

    # svr_severity (from iiapi.h):
    # define IIAPI_SVR_DEFAULT	0x0000
    # define IIAPI_SVR_MESSAGE	0x0001
    # define IIAPI_SVR_WARNING	0x0002
    # define IIAPI_SVR_FORMATTED	0x0010

    _fields_ = [
        ('svr_id_error', II_LONG),
        ('svr_local_error', II_LONG),
        ('svr_id_server', II_LONG),
        ('svr_server_type', II_LONG),
        ('svr_severity', II_LONG),
        ('svr_parmCount', II_INT2),
        ('svr_parmDescr', C.POINTER(IIAPI_DESCRIPTOR)),
        ('svr_parmValue', C.POINTER(IIAPI_DATAVALUE)),
    ]


class IIAPI_II_TRAN_ID(IIAPI_MEMBER):
    '''Ingres transaction ID'''

    _fields_ = [
        ('it_highTran', II_UINT4), 
        ('it_lowTran', II_UINT4)
    ]


class IIAPI_II_DIS_TRAN_ID(IIAPI_MEMBER):
    '''"conventional" 2PC transaction identifier'''

    _fields_ = [
        ('ii_tranID', IIAPI_II_TRAN_ID),
        ('ii_tranName', II_CHAR * IIAPI_TRAN_MAXNAME),
    ]


class IIAPI_XA_TRAN_ID(IIAPI_MEMBER):
    '''XA 2PC transaction ID'''

    _fields_ = [
        ('xt_formatID', II_LONG),
        ('xt_gtridLength', II_LONG),
        ('xt_bqualLength', II_LONG),
        ('xt_data', II_CHAR * IIAPI_XA_XIDDATASIZE),
    ]


class IIAPI_XA_DIS_TRAN_ID(IIAPI_MEMBER):
    '''XA 2PC transaction identifier'''

    _fields_ = [
        ('xa_tranID', IIAPI_XA_TRAN_ID),
        ('xa_branchSeqnum', II_INT4),
        ('xa_branchFlag', II_INT4),
    ]


##  NB this is a union not a struct
class IIAPI_2PC_TRAN_ID(C.Union):
    '''2PC transaction identifier'''

    _fields_ = [
        ('iiXID', IIAPI_II_DIS_TRAN_ID), 
        ('xaXID', IIAPI_XA_DIS_TRAN_ID)
    ]


class IIAPI_TRAN_ID(IIAPI_MEMBER):
    '''
    The IIAPI_TRAN_ID data type specifies and names an OpenAPI
    transaction
    '''

    _fields_ = [
        ('ti_type', II_ULONG), 
        ('ti_value', IIAPI_2PC_TRAN_ID)
    ]


class IIAPI_PARM(IIAPI):
    '''parameter block for the Ingres OpenAPI'''

    ##  I suspect, but don't yet know for sure, that PARM blocks should
    ##  be treated as context managers; they don't do anything (yet) but
    ##  let's make them compliant at least.

    ##  I am having a bit of a crisis of faith over doing this in the
    ##  basic bindings; it should be done in the layer that makes the
    ##  API "python friendly", but I'm leaving it here for now

    def __enter__(self):
        pass

    def __exit__(self):
        pass


class IIAPI_ABORTPARM(IIAPI_PARM):
    '''parameter block for IIapi_abort'''

    _fields_ = [
        ('ab_genParm', IIAPI_GENPARM), 
        ('ab_connHandle', C.c_void_p)
    ]


class IIAPI_AUTOPARM(IIAPI_PARM):
    '''parameter block for IIapi_autocommit'''

    _fields_ = [
        ('ac_genParm', IIAPI_GENPARM),
        ('ac_connHandle', C.c_void_p),
        ('ac_tranHandle', C.c_void_p),
    ]


class IIAPI_BATCHPARM(IIAPI_PARM):
    '''parameter block for IIapi_batch'''

    _fields_ = [
        ('ba_genParm', IIAPI_GENPARM),
        ('ba_connHandle', II_PTR),
        ('ba_queryType', IIAPI_QUERYTYPE),
        ('ba_queryText', II_STR),
        ('ba_parameters', II_BOOL),
        ('ba_flags', II_UINT4),
        ('ba_tranHandle', II_PTR),
        ('ba_stmtHandle', II_PTR),
    ]


class IIAPI_CANCELPARM(IIAPI_PARM):
    '''parameter block for cancelling a query'''

    _fields_ = [('cn_genParm', IIAPI_GENPARM), ('cn_stmtHandle', II_PTR)]


class IIAPI_CATCHEVENTPARM(IIAPI_PARM):
    '''parameter block for IIapi_catchEvent'''

    _fields_ = [
        ('ce_genParm', IIAPI_GENPARM),
        ('ce_connHandle', II_PTR),
        ('ce_selectEventName', II_STR),
        ('ce_selectEventOwner', II_STR),
        ('ce_eventHandle', II_PTR),
        ('ce_eventName', II_STR),
        ('ce_eventOwner', II_STR),
        ('ce_eventDB', II_STR),
        ('ce_eventTime', IIAPI_DATAVALUE),
        ('ce_eventInfoAvail', II_BOOL),
    ]


class IIAPI_CLOSEPARM(IIAPI_PARM):
    '''parameter block for closing a query'''

    _fields_ = [('cl_genParm', IIAPI_GENPARM), ('cl_stmtHandle', II_PTR)]


class IIAPI_COMMITPARM(IIAPI_PARM):
    '''parameter block for IIapi_commit'''

    _fields_ = [('cm_genParm', IIAPI_GENPARM), ('cm_tranHandle', II_PTR)]


class IIAPI_CONNPARM(IIAPI_PARM):
    '''parameter block for IIapi_connect'''

    _fields_ = [
        ('co_genParm', IIAPI_GENPARM),
        ('co_target', II_STR),
        ('co_username', II_STR),
        ('co_password', II_STR),
        ('co_timeout', II_LONG),
        ('co_connHandle', II_PTR),
        ('co_tranHandle', II_PTR),
        ('co_sizeAdvise', II_LONG),
        ('co_apiLevel', II_LONG),
        ('co_type', II_LONG),
    ]


class IIAPI_CONVERTPARM(IIAPI_PARM):
    '''parameter block for IIapi_convertData'''

    _fields_ = [
        ('cv_srcDesc', IIAPI_DESCRIPTOR),
        ('cv_srcValue', IIAPI_DATAVALUE),
        ('cv_dstDesc', IIAPI_DESCRIPTOR),
        ('cv_dstValue', IIAPI_DATAVALUE),
        ('cv_status', IIAPI_STATUS),
    ]


class IIAPI_DISCONNPARM(IIAPI_PARM):
    '''parameter block for disconnecting from a server'''

    _fields_ = [('dc_genParm', IIAPI_GENPARM), ('dc_connHandle', II_PTR)]


class IIAPI_EVENTPARM(IIAPI_PARM):
    '''parameter block for setting up an event handler'''

    _fields_ = [
        ('ev_envHandle', II_PTR),
        ('ev_connHandle', II_PTR),
        ('ev_eventName', II_STR),
        ('ev_eventOwner', II_STR),
        ('ev_eventDB', II_STR),
        ('ev_eventTime', IIAPI_DATAVALUE),
    ]


class IIAPI_FORMATPARM(IIAPI_PARM):
    '''parameter block for IIapi_formatData'''

    _fields_ = [
        ('fd_envHandle', II_PTR),
        ('fd_srcDesc', IIAPI_DESCRIPTOR),
        ('fd_srcValue', IIAPI_DATAVALUE),
        ('fd_dstDesc', IIAPI_DESCRIPTOR),
        ('fd_dstValue', IIAPI_DATAVALUE),
        ('fd_status', IIAPI_STATUS),
    ]


class IIAPI_GETCOLPARM(IIAPI_PARM):
    '''parameter block for IIapi_getColumns'''

    _fields_ = [
        ('gc_genParm', IIAPI_GENPARM),
        ('gc_stmtHandle', II_PTR),
        ('gc_rowCount', II_INT2),  # incorrectly documented II_INT
        ('gc_columnCount', II_INT2),  # incorrectly documented II_INT
        ('gc_columnData', C.POINTER(IIAPI_DATAVALUE)),
        ('gc_rowsReturned', II_INT2),  # incorrectly documented II_INT
        ('gc_moreSegments', II_BOOL),
    ]


class IIAPI_GETCOLINFOPARM(IIAPI_PARM):
    '''parameter block for IIapi_getColumnInfo'''

    _fields_ = [
        ('gi_stmtHandle', II_PTR),
        ('gi_columnNumber', II_INT2),
        ('gi_status', IIAPI_STATUS),
        ('gi_mask', II_ULONG),
        ('gi_lobLength', II_UINT8),
    ]


class IIAPI_GETCOPYMAPPARM(IIAPI_PARM):
    '''parameter block for setting connection parameters'''

    _fields_ = [
        ('gm_genParm', IIAPI_GENPARM),
        ('gm_stmtHandle', II_PTR),
        ('gm_copyMap', IIAPI_COPYMAP),
    ]


class IIAPI_GETDESCRPARM(IIAPI_PARM):
    '''query result descriptor block'''

    _fields_ = [
        ('gd_genParm', IIAPI_GENPARM),
        ('gd_stmtHandle', II_PTR),
        ('gd_descriptorCount', II_INT2),
        ('gd_descriptor', C.POINTER(IIAPI_DESCRIPTOR)),
    ]


class IIAPI_GETEINFOPARM(IIAPI_PARM):
    '''OpenAPI parameter block for returning error and user-defined info'''

    _fields_ = [
        ('ge_errorHandle', II_PTR),
        ('ge_type', II_LONG),
        ('ge_SQLSTATE', II_CHAR * (II_SQLSTATE_LEN + 1)),
        ('ge_errorCode', II_LONG),
        ('ge_message', II_STR),
        ('ge_serverInfoAvail', II_BOOL),
        ('ge_serverInfo', C.POINTER(IIAPI_SVR_ERRINFO)),
        ('ge_status', IIAPI_STATUS),
    ]


class IIAPI_GETEVENTPARM(IIAPI_PARM):
    '''parameter block for IIapi_getEvent'''

    _fields_ = [
        ('gv_genParm', IIAPI_GENPARM),
        ('gv_connHandle', C.c_void_p),
        ('gv_timeout', C.c_int),
    ]


class IIAPI_GETQINFOPARM(IIAPI_PARM):
    '''parameter block for IIapi_getQueryInfo()'''

    _fields_ = [
        ('gq_genParm', IIAPI_GENPARM),
        ('gq_stmtHandle', II_PTR),
        ('gq_flags', II_ULONG),
        ('gq_mask', II_ULONG),
        ('gq_rowCount', II_LONG),
        ('gq_readonly', II_BOOL),
        ('gq_procedureReturn', II_LONG),
        ('gq_procedureHandle', II_PTR),
        ('gq_repeatQueryHandle', II_PTR),
        ('gq_tableKey', II_CHAR * IIAPI_TBLKEYSZ),
        ('gq_objectKey', II_CHAR * IIAPI_OBJKEYSZ),
        ('gq_cursorType', II_ULONG),
        ('gq_rowStatus', II_ULONG),
        (
            'gq_rowPosition',
            II_LONG,
        ),
        ('gq_rowCountEx', II_INT8),
    ]


class IIAPI_INITPARM(IIAPI_PARM):
    '''parameter block for initializing the Ingres OpenAPI'''

    _fields_ = [
        ('in_timeout', II_LONG),
        ('in_version', II_LONG),
        ('in_status', IIAPI_STATUS),
        ('in_envHandle', II_PTR),
    ]


class IIAPI_MODCONNPARM(IIAPI_PARM):
    '''parameter block for IIapi_modifyConnect()'''

    _fields_ = [('mc_genParm', IIAPI_GENPARM), ('mc_connHandle', II_PTR)]


class IIAPI_POSPARM(IIAPI_PARM):
    '''parameter block for IIapi_position()'''

    _fields_ = [
        ('po_genParm', IIAPI_GENPARM),
        ('po_stmtHandle', II_PTR),
        ('po_reference', II_UINT2),
        ('po_offset', II_INT),
        ('po_rowCount', II_INT2),
    ]


class IIAPI_PREPCMTPARM(IIAPI_PARM):
    '''parameter block for IIapi_prepareCommit()'''

    _fields_ = [('pr_genParm', IIAPI_GENPARM), ('pr_tranHandle', II_PTR)]


class IIAPI_PUTCOLPARM(IIAPI_PARM):
    '''parameter block for IIapi_putColumns()'''

    _fields_ = [
        ('pc_genParm', IIAPI_GENPARM),
        ('pc_stmtHandle', II_PTR),
        ('pc_columnCount', II_INT2),
        ('pc_columnData', C.POINTER(IIAPI_DATAVALUE)),
        ('pc_moreSegments', II_BOOL),
    ]


class IIAPI_PUTPARMPARM(IIAPI_PARM):
    '''parameter block for IIapi_putParms()'''

    _fields_ = [
        ('pp_genParm', IIAPI_GENPARM),
        ('pp_stmtHandle', II_PTR),
        ('pp_parmCount', II_INT2),
        ('pp_parmData', C.POINTER(IIAPI_DATAVALUE)),
        ('pp_moreSegments', II_BOOL),
    ]


class IIAPI_QUERYPARM(IIAPI_PARM):
    '''parameter block for IIapi_query()'''

    _fields_ = [
        ('qy_genParm', IIAPI_GENPARM),
        ('qy_connHandle', II_PTR),
        ('qy_queryType', IIAPI_QUERYTYPE),
        ('qy_queryText', II_STR),
        ('qy_parameters', II_BOOL),
        ('qy_tranHandle', II_PTR),
        ('qy_stmtHandle', II_PTR),
        ('qy_flags', II_ULONG),
    ]


class IIAPI_REGXIDPARM(IIAPI_PARM):
    '''parameter block for IIapi_registerXID()'''

    _fields_ = [
        ('rg_tranID', IIAPI_TRAN_ID),
        ('rg_tranIdHandle', II_PTR),
        ('rg_status', IIAPI_STATUS),
    ]


class IIAPI_RELENVPARM(IIAPI_PARM):
    '''parameter block for IIapi_releaseEnv()'''

    _fields_ = [('re_envHandle', II_PTR), ('re_status', IIAPI_STATUS)]


class IIAPI_RELXIDPARM(IIAPI_PARM):
    '''parameter block for IIapi_releaseXID()'''

    _fields_ = [('rl_tranIdHandle', II_PTR), ('rl_status', IIAPI_STATUS)]


class IIAPI_ROLLBACKPARM(IIAPI_PARM):
    '''parameter block for IIapi_rollback()'''

    _fields_ = [
        ('rb_genParm', IIAPI_GENPARM),
        ('rb_tranHandle', II_PTR),
        ('rb_savePointHandle', II_PTR),
    ]


class IIAPI_SAVEPTPARM(IIAPI_PARM):
    '''parameter block for IIapi_savePoint()'''

    _fields_ = [
        ('sp_genParm', IIAPI_GENPARM),
        ('sp_tranHandle', II_PTR),
        ('sp_savePoint', II_STR),
        ('sp_savePointHandle', II_PTR),
    ]


class IIAPI_SCROLLPARM(IIAPI_PARM):
    '''parameter block for IIapi_scroll()'''

    _fields_ = [
        ('sl_genParm', IIAPI_GENPARM),
        ('sl_stmtHandle', II_PTR),
        ('sl_orientation', II_UINT2),
        ('sl_offset', II_INT),
    ]


class IIAPI_SETCONPRMPARM(IIAPI_PARM):
    '''parameter block for setConnectParam()'''

    _fields_ = [
        ('sc_genParm', IIAPI_GENPARM),
        ('sc_connHandle', C.c_void_p),
        ('sc_paramID', C.c_int),
        ('sc_paramValue', C.c_void_p),
    ]


class IIAPI_SETDESCRPARM(IIAPI_PARM):
    '''parameter block for IIapi_setDescriptor()'''

    _fields_ = [
        ('sd_genParm', IIAPI_GENPARM),
        ('sd_stmtHandle', II_PTR),
        ('sd_descriptorCount', II_INT2),
        ('sd_descriptor', C.POINTER(IIAPI_DESCRIPTOR)),
    ]


class IIAPI_SETENVPRMPARM(IIAPI_PARM):
    '''parameter block for IIapi_setEnvParam()'''

    _fields_ = [
        ('se_envHandle', C.c_void_p),
        ('se_paramID', C.c_int),
        ('se_paramValue', C.c_void_p),
        ('se_status', C.c_uint),
    ]


class IIAPI_TERMPARM(IIAPI_PARM):
    '''parameter block for terminating the Ingres OpenAPI'''

    _fields_ = [('tm_status', IIAPI_STATUS)]


class IIAPI_TRACEPARM(IIAPI_PARM):
    '''parameter block to set up a trace callback'''

    _fields_ = [
        ('tr_envHandle', II_PTR),
        ('tr_connHandle', II_PTR),
        ('tr_length', II_INT4),
        ('tr_message', II_STR),
    ]


class IIAPI_WAITPARM(IIAPI_PARM):
    '''parameter block for waiting'''

    _fields_ = [('wt_timeout', II_LONG), ('wt_status', IIAPI_STATUS)]


class IIAPI_XACOMMITPARM(IIAPI_PARM):
    '''parameter block for IIapi_xaCommit()'''

    _fields_ = [
        ('xc_genParm', IIAPI_GENPARM),
        ('xc_connHandle', II_PTR),
        ('xc_tranID', IIAPI_TRAN_ID),
        ('xc_flags', II_ULONG),
    ]


class IIAPI_XAENDPARM(IIAPI_PARM):
    '''parameter block for IIapi_xaEnd()'''

    _fields_ = [
        ('xe_genParm', IIAPI_GENPARM),
        ('xe_connHandle', II_PTR),
        ('xe_tranID', IIAPI_TRAN_ID),
        ('xe_flags', II_ULONG),
    ]


class IIAPI_XAPREPPARM(IIAPI_PARM):
    '''parameter block for IIapi_xaPrepare()'''

    _fields_ = [
        ('xp_genParm', IIAPI_GENPARM),
        ('xp_connHandle', II_PTR),
        ('xp_tranID', IIAPI_TRAN_ID),
        ('xp_flags', II_ULONG),
    ]


class IIAPI_XAROLLPARM(IIAPI_PARM):
    '''parameter block for IIapi_xaRollback()'''

    _fields_ = [
        ('xr_genParm', IIAPI_GENPARM),
        ('xr_connHandle', II_PTR),
        ('xr_tranID', IIAPI_TRAN_ID),
        ('xr_flags', II_ULONG),
    ]


class IIAPI_XASTARTPARM(IIAPI_PARM):
    '''parameter block for IIapi_xaStart()'''

    _fields_ = [
        ('xs_genParm', IIAPI_GENPARM),
        ('xs_connHandle', II_PTR),
        ('xs_tranID', IIAPI_TRAN_ID),
        ('xs_flags', II_ULONG),
        ('xs_tranHandle', II_PTR),
    ]



##  Copyright (c) 2024 Rational Commerce Ltd.

'''
non-blocking Python asyncio-aware Actian Vector/X/Ingres OpenAPI
'''


from functools import wraps
from loguru import logger
import asyncio
import pyngres as py
from .IIAPI_CONSTANTS import *
from .IIAPI_PARM import *
##  the following functions are synchronous and not awaitable; we just
##  import them into the pyngres.asyncio namespace
from pyngres import ( IIapi_convertData, IIapi_formatData, IIapi_getColumnInfo, 
    IIapi_getErrorInfo, IIapi_initialize, IIapi_registerXID, IIapi_releaseEnv, 
    IIapi_releaseXID, IIapi_setEnvParam, IIapi_terminate ) 
##  import the callback helper functions
from pyngres import ( IIapi_callback, IIapi_getCallbackPtr,
    IIapi_getClosurePtr, IIapi_getClosure ) 


##  the IIAPI_DEV_MODE envar puts pyngres.asyncio in development mode
_name = __name__
if 'IIAPI_DEV_MODE' in os.environ:
    IIAPI_DEV_MODE = True
    logger.warning(f'using {_name} in development mode')
    logger.info(f'to disable logging messages: logger.disable(\'{_name}\')')
    logger.info(f'to enable logging messages: logger.enable(\'{_name}\')')
else:
    IIAPI_DEV_MODE = False
    logger.disable(_name)


_NOWAIT = py.IIAPI_WAITPARM()
_NOWAIT.wt_timeout = -1


def _IIapi_yield_to_OpenAPI( future ):
    py.IIapi_wait( _NOWAIT )
    future.set_result(True)
    return future


async def _IIapi_complete( future ):
    await asyncio.sleep(0.0)
    _IIapi_yield_to_OpenAPI( future )
    await future


def _IIapi_awaitable( IIapi_function ):
    '''make a pyngres function awaitable'''
    @wraps(IIapi_function)
    async def awaitable( pcb ):
        loop = asyncio.get_event_loop()
        genParm = pcb.genParm()
        IIapi_function( pcb )
        ##  poll for completion of the the OpenAPI
        while not genParm.gp_completed:
            future = loop.create_future()
            await _IIapi_complete( future) 
    return awaitable


@_IIapi_awaitable
def IIapi_abort( pcb ):
    py.IIapi_abort( pcb )


@_IIapi_awaitable
def IIapi_autocommit(pcb):
	py.IIapi_autocommit( pcb )


@_IIapi_awaitable
def IIapi_batch(pcb):
	py.IIapi_batch( pcb )


@_IIapi_awaitable
def IIapi_cancel(pcb):
	py.IIapi_cancel( pcb )


@_IIapi_awaitable
def IIapi_catchEvent(pcb):
	py.IIapi_catchEvent( pcb )


@_IIapi_awaitable
def IIapi_close(pcb):
	py.IIapi_close( pcb )


@_IIapi_awaitable
def IIapi_commit(pcb):
	py.IIapi_commit( pcb )


@_IIapi_awaitable
def IIapi_connect(pcb):
	py.IIapi_connect( pcb )


@_IIapi_awaitable
def IIapi_disconnect(pcb):
	py.IIapi_disconnect( pcb )


@_IIapi_awaitable
def IIapi_getColumns(pcb):
	py.IIapi_getColumns( pcb )


@_IIapi_awaitable
def IIapi_getCopyMap(pcb):
	py.IIapi_getCopyMap( pcb )


@_IIapi_awaitable
def IIapi_getDescriptor(pcb):
	py.IIapi_getDescriptor( pcb )


@_IIapi_awaitable
def IIapi_getEvent(pcb):
	py.IIapi_getEvent( pcb )


@_IIapi_awaitable
def IIapi_getQueryInfo(pcb):
	py.IIapi_getQueryInfo( pcb )


@_IIapi_awaitable
def IIapi_modifyConnect(pcb):
	py.IIapi_modifyConnect( pcb )


@_IIapi_awaitable
def IIapi_position(pcb):
	py.IIapi_position( pcb )


@_IIapi_awaitable
def IIapi_prepareCommit(pcb):
	py.IIapi_prepareCommit( pcb )


@_IIapi_awaitable
def IIapi_putColumns(pcb):
	py.IIapi_putColumns( pcb )


@_IIapi_awaitable
def IIapi_putParms(pcb):
	py.IIapi_putParms( pcb )


@_IIapi_awaitable
def IIapi_query(pcb):
	py.IIapi_query( pcb )


@_IIapi_awaitable
def IIapi_rollback(pcb):
	py.IIapi_rollback( pcb )


@_IIapi_awaitable
def IIapi_savePoint(pcb):
	py.IIapi_savePoint( pcb )


@_IIapi_awaitable
def IIapi_scroll(pcb):
	py.IIapi_scroll( pcb )


@_IIapi_awaitable
def IIapi_setConnectParam(pcb):
	py.IIapi_setConnectParam( pcb )


@_IIapi_awaitable
def IIapi_setDescriptor(pcb):
	py.IIapi_setDescriptor( pcb )


##  pyngres.asyncio makes IIapi_wait() defunct so don't add it. (If someone
##  thinks they have a use for it they can explicitly import it from pyngres)


@_IIapi_awaitable
def IIapi_xaCommit(pcb):
	py.IIapi_xaCommit( pcb )


@_IIapi_awaitable
def IIapi_xaStart(pcb):
	py.IIapi_xaStart( pcb )


@_IIapi_awaitable
def IIapi_xaEnd(pcb):
	py.IIapi_xaEnd( pcb )


@_IIapi_awaitable
def IIapi_xaPrepare(pcb):
	py.IIapi_xaPrepare( pcb )


@_IIapi_awaitable
def IIapi_xaRollback(pcb):
	py.IIapi_xaRollback( pcb )



##  all of these values are harvested from $II_SYSTEM/ingres/files/iiapi.h,
##  so they will have to be repeatedly reviewed for breaking changes

IIAPI_VERSION_1 = 1
IIAPI_VERSION_2 = 2
IIAPI_VERSION_3 = 3
IIAPI_VERSION_4 = 4
IIAPI_VERSION_5 = 5
IIAPI_VERSION_6 = 6
IIAPI_VERSION_7 = 7
IIAPI_VERSION_8 = 8
IIAPI_VERSION_9 = 9
IIAPI_VERSION_10 = 10
IIAPI_VERSION_11 = 11
IIAPI_VERSION = IIAPI_VERSION_11

IIAPI_LEVEL_0 = 0
IIAPI_LEVEL_1 = 1
IIAPI_LEVEL_2 = 2
IIAPI_LEVEL_3 = 3
IIAPI_LEVEL_4 = 4
IIAPI_LEVEL_5 = 5
IIAPI_LEVEL_6 = 6
IIAPI_LEVEL_7 = 7

IIAPI_HNDL_TYPE = -1       ## API Handle 
IIAPI_CHR_TYPE = 32        ## Text Fixed 
IIAPI_CHA_TYPE = 20        ## Character Fixed 
IIAPI_VCH_TYPE = 21        ## Character Variable 
IIAPI_LVCH_TYPE = 22       ## Character Long 
IIAPI_LCLOC_TYPE = 36      ## Character Long Locator 
IIAPI_BYTE_TYPE = 23       ## Binary Fixed 
IIAPI_VBYTE_TYPE = 24      ## Binary Variable 
IIAPI_LBYTE_TYPE = 25      ## Binary Long 
IIAPI_LBLOC_TYPE = 35      ## Binary Long Locator 
IIAPI_NCHA_TYPE = 26       ## Unicode Fixed 
IIAPI_NVCH_TYPE = 27       ## Unicode Variable 
IIAPI_LNVCH_TYPE = 28      ## Unicode Long 
IIAPI_LNLOC_TYPE = 29      ## Unicode Long Locator 
IIAPI_TXT_TYPE = 37        ## Text Variable 
IIAPI_LTXT_TYPE = 41       ## Text Var (typeless NULL) 
IIAPI_INT_TYPE = 30        ## Integer 
IIAPI_FLT_TYPE = 31        ## Floating Point 
IIAPI_MNY_TYPE =  5        ## Money 
IIAPI_DEC_TYPE = 10        ## Decimal 
IIAPI_BOOL_TYPE = 38       ## Boolean 
IIAPI_UUID_TYPE = 55       ## UUID 
IIAPI_IPV4_TYPE = 66       ## IPv4 
IIAPI_IPV6_TYPE = 67       ## IPv6 
IIAPI_DTE_TYPE =  3        ## Ingres Date 
IIAPI_IDATE_TYPE = IIAPI_DTE_TYPE
IIAPI_DATE_TYPE =  4       ## ANSI Date 
IIAPI_ADATE_TYPE = IIAPI_DATE_TYPE
IIAPI_TIME_TYPE =  8       ## Time 
IIAPI_TMWO_TYPE =  6       ## Time without Timezone 
IIAPI_TMTZ_TYPE =  7       ## Time with Timezone 
IIAPI_TS_TYPE = 19         ## Timestamp 
IIAPI_TSWO_TYPE =  9       ## Timestamp without Timezone 
IIAPI_TSTZ_TYPE = 18       ## Timestamp with Timezone 
IIAPI_INTYM_TYPE = 33      ## Interval Year to Month 
IIAPI_INTDS_TYPE = 34      ## Interval Day to Second 
IIAPI_LOGKEY_TYPE = 11     ## Logical Key 
IIAPI_TABKEY_TYPE = 12     ## Table Key 
IIAPI_GEOM_TYPE = 56       ## Spatial Long Byte 
IIAPI_POINT_TYPE = 57      ## Point Long Byte 
IIAPI_MPOINT_TYPE = 58     ## MPoint Long Byte 
IIAPI_LINE_TYPE = 59       ## Line Long Byte 
IIAPI_MLINE_TYPE = 61      ## MLine Long Byte 
IIAPI_POLY_TYPE = 62       ## Poly Long Byte 
IIAPI_MPOLY_TYPE = 63      ## Mpoly Long Byte 
IIAPI_NBR_TYPE = 64        ## NBR Byte 
IIAPI_GEOMC_TYPE = 65      ## Geomc Long Byte 
IIAPI_CURVE_TYPE = 68      ## CURVE Spatial Type 
IIAPI_SURF_TYPE = 69       ## SURF Spatial Type 
IIAPI_PSURF_TYPE = 70      ## PSURF Spatial Type 
IIAPI_GEOMZ_TYPE = 71      ## GEOMZ Spatial Type 
IIAPI_POINTZ_TYPE = 72     ## POINTZ Spatial Type 
IIAPI_LINEZ_TYPE = 73      ## LINEZ Spatial Type 
IIAPI_POLYZ_TYPE = 74      ## POLYZ Spatial Type 
IIAPI_MPOINTZ_TYPE = 75    ## MPOINTZ Spatial Type 
IIAPI_MLINEZ_TYPE = 76     ## MLINEZ Spatial Type 
IIAPI_MPOLYZ_TYPE = 77     ## MPOLYZ Spatial Type 
IIAPI_GEOMCZ_TYPE = 78     ## GEOMCZ Spatial Type 
IIAPI_CURVEZ_TYPE = 79     ## CURVEZ Spatial Type 
IIAPI_SURFZ_TYPE = 80      ## SURFZ Spatial Type 
IIAPI_PSURFZ_TYPE = 81     ## PSURFZ Spatial Type 
IIAPI_GEOMM_TYPE = 82      ## GEOMM Spatial Type 
IIAPI_POINTM_TYPE = 83     ## POINTM Spatial Type 
IIAPI_LINEM_TYPE = 84      ## LINEM Spatial Type 
IIAPI_POLYM_TYPE = 85      ## POLYM Spatial Type 
IIAPI_MPOINTM_TYPE = 86    ## MPOINTM Spatial Type 
IIAPI_MLINEM_TYPE = 87     ## MLINEM Spatial Type 
IIAPI_MPOLYM_TYPE = 88     ## MPOLYM Spatial Type 
IIAPI_GEOMCM_TYPE = 89     ## GEOMCM Spatial Type 
IIAPI_CURVEM_TYPE = 90     ## CURVEM Spatial Type 
IIAPI_SURFM_TYPE = 91      ## SURFM Spatial Type 
IIAPI_PSURFM_TYPE = 92     ## PSURFM Spatial Type 
IIAPI_GEOMZM_TYPE = 93     ## GEOMZM Spatial Type 
IIAPI_POINTZM_TYPE = 94    ## POINTZM Spatial Type 
IIAPI_LINEZM_TYPE = 95     ## LINEZM Spatial Type 
IIAPI_POLYZM_TYPE = 100    ## POLYZM Spatial Type 
IIAPI_MPOINTZM_TYPE = 101  ## MPOINTZM Spatial Type 
IIAPI_MLINEZM_TYPE = 102   ## MLINEZM Spatial Type 
IIAPI_MPOLYZM_TYPE = 103   ## MPOLYZM Spatial Type 
IIAPI_GEOMCZM_TYPE = 104   ## GEOMCZM Spatial Type 
IIAPI_CURVEZM_TYPE = 105   ## CURVEZM Spatial Type 
IIAPI_SURFZM_TYPE = 106    ## SURFZM Spatial Type 
IIAPI_PSURFZM_TYPE = 107   ## PSURFZM Spatial Type 
IIAPI_MCURVE_TYPE = 108    ## MCURVE Spatial Type 
IIAPI_MSURF_TYPE = 109     ## MSURF Spatial Type 
IIAPI_MCURVEZ_TYPE = 110   ## MCURVEZ Spatial Type 
IIAPI_MSURFZ_TYPE = 111    ## MSURFZ Spatial Type 
IIAPI_MCURVEM_TYPE = 112   ## MCURVEM Spatial Type 
IIAPI_MSURFM_TYPE = 113    ## MSURFM Spatial Type 
IIAPI_MCURVEZM_TYPE = 114  ## MCURVEZM Spatial Type 
IIAPI_MSURFZM_TYPE = 115   ## MSURFZM Spatial Type 
IIAPI_CSTR_TYPE = 116      ## CSTR Spatial Type 
IIAPI_CSTRZ_TYPE = 117     ## CSTRZ Spatial Type 
IIAPI_CSTRM_TYPE = 118     ## CSTRM Spatial Type 
IIAPI_CSTRZM_TYPE = 119    ## CSTRZM Spatial Type 
IIAPI_CCURVE_TYPE = 120    ## CCURVE Spatial Type 
IIAPI_CCURVEZ_TYPE = 121   ## CCURVEZ Spatial Type 
IIAPI_CCURVEM_TYPE = 122   ## CCURVEM Spatial Type 
IIAPI_CCURVEZM_TYPE = 123  ## CCURVEZM Spatial Type 
IIAPI_CPOLY_TYPE = 124     ## CPOLY Spatial Type 
IIAPI_CPOLYZ_TYPE = 125    ## CPOLYZ Spatial Type 
IIAPI_CPOLYM_TYPE = 126    ## CPOLYM Spatial Type 
IIAPI_CPOLYZM_TYPE = 127   ## CPOLYZM Spatial Type 
IIAPI_TIN_TYPE = 128       ## TIN Spatial Type 
IIAPI_TINZ_TYPE = 129      ## TINZ Spatial Type 
IIAPI_TINM_TYPE = 130      ## TINM Spatial Type 
IIAPI_TINZM_TYPE = 131     ## TINZM Spatial Type 
IIAPI_TRI_TYPE = 132       ## TRI Spatial Type 
IIAPI_TRIZ_TYPE = 133      ## TRIZ Spatial Type 
IIAPI_TRIM_TYPE = 134      ## TRIM Spatial Type 
IIAPI_TRIZM_TYPE = 135     ## TRIZM Spatial Type 
IIAPI_BOX_TYPE = 136       ## BOX Spatial Type 
IIAPI_BOXZ_TYPE = 137      ## BOXZ Spatial Type 

IIAPI_VAR_TYPES = {IIAPI_VBYTE_TYPE, IIAPI_VCH_TYPE, IIAPI_NVCH_TYPE}
IIAPI_LONG_TYPES = {IIAPI_LVCH_TYPE, IIAPI_LBYTE_TYPE, IIAPI_LNVCH_TYPE, 
    IIAPI_GEOM_TYPE, IIAPI_POINT_TYPE, IIAPI_MPOINT_TYPE,
    IIAPI_LINE_TYPE, IIAPI_MLINE_TYPE, IIAPI_POLY_TYPE,
    IIAPI_MPOLY_TYPE, IIAPI_NBR_TYPE, IIAPI_GEOMC_TYPE}


IIAPI_BOOL_LEN = 1
IIAPI_DATE_LEN = 4
IIAPI_ADATE_LEN = IIAPI_DATE_LEN
IIAPI_DTE_LEN = 12
IIAPI_EP_MAX_SEGMENT_LEN = 24
IIAPI_F4_LEN = 4
IIAPI_F8_LEN = 8
IIAPI_GI_LOB_LENGTH = 0x0001
IIAPI_I1_LEN = 1
IIAPI_I2_LEN = 2
IIAPI_I4_LEN = 4
IIAPI_I8_LEN = 8
IIAPI_IDATE_LEN = IIAPI_DTE_LEN
IIAPI_INTDS_LEN = 12
IIAPI_INTYM_LEN = 3
IIAPI_LOCATOR_LEN = 4
IIAPI_LOGKEY_LEN = 16
IIAPI_MNY_LEN = 8
IIAPI_SEGMENT_LEN = 2000
IIAPI_TABKEY_LEN = 8
IIAPI_TIME_LEN = 10
IIAPI_TS_LEN = 14
IIAPI_OBJKEYSZ = 16
IIAPI_TBLKEYSZ = 8
IIAPI_UUID_LEN = 16
IIAPI_IPV4_LEN = 4
IIAPI_IPV6_LEN = 16


IIAPI_MIB_CONN = 'exp.aif.api.conn'
IIAPI_MIB_CONN_GCA_ASSOC = 'exp.aif.api.conn.gca_assoc'
IIAPI_MIB_CONN_TARGET = 'exp.aif.api.conn.target'
IIAPI_MIB_CONN_USERID = 'exp.aif.api.conn.userid'
IIAPI_MIB_TRACE_LEVEL = 'exp.aif.api.trace_level'

IIAPI_ST_SUCCESS = 0
IIAPI_ST_MESSAGE = 1
IIAPI_ST_WARNING = 2
IIAPI_ST_NO_DATA = 3
IIAPI_ST_ERROR = 4
IIAPI_ST_FAILURE = 5
IIAPI_ST_NOT_INITIALIZED = 6
IIAPI_ST_INVALID_HANDLE = 7
IIAPI_ST_OUT_OF_MEMORY = 8
IIAPI_ST_MSG = {
    IIAPI_ST_SUCCESS: 'IIAPI_ST_SUCCESS',
    IIAPI_ST_MESSAGE: 'IIAPI_ST_MESSAGE',
    IIAPI_ST_WARNING: 'IIAPI_ST_WARNING',
    IIAPI_ST_NO_DATA: 'IIAPI_ST_NO_DATA',
    IIAPI_ST_ERROR: 'IIAPI_ST_ERROR',
    IIAPI_ST_FAILURE: 'IIAPI_ST_FAILURE',
    IIAPI_ST_NOT_INITIALIZED: 'IIAPI_ST_NOT_INITIALIZED',
    IIAPI_ST_INVALID_HANDLE: 'IIAPI_ST_INVALID_HANDLE',
    IIAPI_ST_OUT_OF_MEMORY: 'IIAPI_ST_OUT_OF_MEMORY',
}

IIAPI_CURSOR_SCROLL = 2
IIAPI_CURSOR_UPDATE = 1

IIAPI_COL_PROCBYREFPARM = 1
IIAPI_COL_PROCGTTPARM = 5
IIAPI_COL_PROCINOUTPARM = 1
IIAPI_COL_PROCINPARM = 2
IIAPI_COL_PROCOUTPARM = 6
IIAPI_COL_PROCPARM = 2
IIAPI_COL_QPARM = 4
IIAPI_COL_SVCPARM = 3
IIAPI_COL_TUPLE = 0

IIAPI_CP_APP_ID = 27
IIAPI_CP_APPLICATION = 26
IIAPI_CP_CENTURY_BOUNDARY = 37
IIAPI_CP_CHAR_WIDTH = 1
IIAPI_CP_BASE = IIAPI_CP_CHAR_WIDTH
IIAPI_CP_DATE_ALIAS = 40
IIAPI_CP_DATE_FORMAT = 19
IIAPI_CP_DBMS_PASSWORD = 36
IIAPI_CP_DECIMAL_CHAR = 16
IIAPI_CP_EFFECTIVE_USER = 29
IIAPI_CP_EXCLUSIVE_LOCK = 32
IIAPI_CP_EXCLUSIVE_SYS_UPDATE = 30
IIAPI_CP_FLOAT4_PRECISION = 8
IIAPI_CP_FLOAT4_STYLE = 11
IIAPI_CP_FLOAT4_WIDTH = 6
IIAPI_CP_FLOAT8_PRECISION = 9
IIAPI_CP_FLOAT8_STYLE = 12
IIAPI_CP_FLOAT8_WIDTH = 7
IIAPI_CP_GATEWAY_PARM = 35
IIAPI_CP_GROUP_ID = 28
IIAPI_CP_INT1_WIDTH = 3
IIAPI_CP_INT2_WIDTH = 4
IIAPI_CP_INT4_WIDTH = 5
IIAPI_CP_INT8_WIDTH = 39
IIAPI_CP_LOGIN_LOCAL = 38
IIAPI_CP_MATH_EXCP = 17
IIAPI_CP_MISC_PARM = 34
IIAPI_CP_MONEY_LORT = 15
IIAPI_CP_MONEY_PRECISION = 10
IIAPI_CP_MONEY_SIGN = 14
IIAPI_CP_NATIVE_LANG = 24
IIAPI_CP_NATIVE_LANG_CODE = 25
IIAPI_CP_NUMERIC_TREATMENT = 13
IIAPI_CP_RESULT_TBL = 22
IIAPI_CP_SECONDARY_INX = 21
IIAPI_CP_SERVER_TYPE = 23
IIAPI_CP_SHARED_SYS_UPDATE = 31
IIAPI_CP_STRING_TRUNC = 18
IIAPI_CP_TIMEZONE = 20
IIAPI_CP_TOP = IIAPI_CP_DATE_ALIAS
IIAPI_CP_TXT_WIDTH = 2
IIAPI_CP_WAIT_LOCK = 33

IIAPI_CT_NS = 3
IIAPI_CT_SQL = 1

IIAPI_EP_CENTURY_BOUNDARY = 23
IIAPI_EP_CHAR_WIDTH = 1
IIAPI_EP_BASE = IIAPI_EP_CHAR_WIDTH
IIAPI_EP_DATE_ALIAS = 29
IIAPI_EP_DATE_FORMAT = 19
IIAPI_EP_DECIMAL_CHAR = 16
IIAPI_EP_EVENT_FUNC = 27
IIAPI_EP_FLOAT4_PRECISION = 8
IIAPI_EP_FLOAT4_STYLE = 11
IIAPI_EP_FLOAT4_WIDTH = 6
IIAPI_EP_FLOAT8_PRECISION = 9
IIAPI_EP_FLOAT8_STYLE = 12
IIAPI_EP_FLOAT8_WIDTH = 7
IIAPI_EP_INT1_WIDTH = 3
IIAPI_EP_INT2_WIDTH = 4
IIAPI_EP_INT4_WIDTH = 5
IIAPI_EP_INT8_WIDTH = 28
IIAPI_EP_MATH_EXCP = 17
IIAPI_EP_MONEY_LORT = 15
IIAPI_EP_MONEY_PRECISION = 10
IIAPI_EP_MONEY_SIGN = 14
IIAPI_EP_NATIVE_LANG = 21
IIAPI_EP_NATIVE_LANG_CODE = 22
IIAPI_EP_NUMERIC_TREATMENT = 13
IIAPI_EP_PROMPT_FUNC = 26
IIAPI_EP_STRING_TRUNC = 18
IIAPI_EP_TIMEZONE = 20
IIAPI_EP_TOP = IIAPI_EP_DATE_ALIAS
IIAPI_EP_TRACE_FUNC = 25
IIAPI_EP_TXT_WIDTH = 2
IIAPI_EP_MSG = {
    IIAPI_EP_CENTURY_BOUNDARY: 'IIAPI_EP_CENTURY_BOUNDARY',
    IIAPI_EP_CHAR_WIDTH: 'IIAPI_EP_CHAR_WIDTH',
    IIAPI_EP_BASE: 'IIAPI_EP_BASE',
    IIAPI_EP_DATE_ALIAS: 'IIAPI_EP_DATE_ALIAS',
    IIAPI_EP_DATE_FORMAT: 'IIAPI_EP_DATE_FORMAT',
    IIAPI_EP_DECIMAL_CHAR: 'IIAPI_EP_DECIMAL_CHAR',
    IIAPI_EP_EVENT_FUNC: 'IIAPI_EP_EVENT_FUNC',
    IIAPI_EP_FLOAT4_PRECISION: 'IIAPI_EP_FLOAT4_PRECISION',
    IIAPI_EP_FLOAT4_STYLE: 'IIAPI_EP_FLOAT4_STYLE',
    IIAPI_EP_FLOAT4_WIDTH: 'IIAPI_EP_FLOAT4_WIDTH',
    IIAPI_EP_FLOAT8_PRECISION: 'IIAPI_EP_FLOAT8_PRECISION',
    IIAPI_EP_FLOAT8_STYLE: 'IIAPI_EP_FLOAT8_STYLE',
    IIAPI_EP_FLOAT8_WIDTH: 'IIAPI_EP_FLOAT8_WIDTH',
    IIAPI_EP_INT1_WIDTH: 'IIAPI_EP_INT1_WIDTH',
    IIAPI_EP_INT2_WIDTH: 'IIAPI_EP_INT2_WIDTH',
    IIAPI_EP_INT4_WIDTH: 'IIAPI_EP_INT4_WIDTH',
    IIAPI_EP_INT8_WIDTH: 'IIAPI_EP_INT8_WIDTH',
    IIAPI_EP_MATH_EXCP: 'IIAPI_EP_MATH_EXCP',
    IIAPI_EP_MONEY_LORT: 'IIAPI_EP_MONEY_LORT',
    IIAPI_EP_MONEY_PRECISION: 'IIAPI_EP_MONEY_PRECISION',
    IIAPI_EP_MONEY_SIGN: 'IIAPI_EP_MONEY_SIGN',
    IIAPI_EP_NATIVE_LANG: 'IIAPI_EP_NATIVE_LANG',
    IIAPI_EP_NATIVE_LANG_CODE: 'IIAPI_EP_NATIVE_LANG_CODE',
    IIAPI_EP_NUMERIC_TREATMENT: 'IIAPI_EP_NUMERIC_TREATMENT',
    IIAPI_EP_PROMPT_FUNC: 'IIAPI_EP_PROMPT_FUNC',
    IIAPI_EP_STRING_TRUNC: 'IIAPI_EP_STRING_TRUNC',
    IIAPI_EP_TIMEZONE: 'IIAPI_EP_TIMEZONE',
    IIAPI_EP_TOP: 'IIAPI_EP_TOP',
    IIAPI_EP_TRACE_FUNC: 'IIAPI_EP_TRACE_FUNC',
    IIAPI_EP_TXT_WIDTH: 'IIAPI_EP_TXT_WIDTH',
}


IIAPI_EPV_ANSIDATE = 'ansidate'
IIAPI_EPV_DECASDEC = 'd'
IIAPI_EPV_DECASFLOAT = 'f'
IIAPI_EPV_DFRMT_ANSI = 10
IIAPI_EPV_DFRMT_DMY = 7
IIAPI_EPV_DFRMT_FINNISH = 2
IIAPI_EPV_DFRMT_GERMAN = 4
IIAPI_EPV_DFRMT_ISO = 3
IIAPI_EPV_DFRMT_ISO4 = 9
IIAPI_EPV_DFRMT_ISO4T = 11
IIAPI_EPV_DFRMT_ISO4TC = 12
IIAPI_EPV_DFRMT_ISO8601 = 13
IIAPI_EPV_DFRMT_MDY = 6
IIAPI_EPV_DFRMT_MLT4 = 8
IIAPI_EPV_DFRMT_MULTI = 1
IIAPI_EPV_DFRMT_US = 0
IIAPI_EPV_DFRMT_YMD = 5
IIAPI_EPV_EXPONENTIAL = 'e'
IIAPI_EPV_FLOATDEC = 'n'
IIAPI_EPV_FLOATDECALIGN = 'g'
IIAPI_EPV_FLOATF = 'f'
IIAPI_EPV_INGDATE = 'ingresdate'
IIAPI_EPV_MONEY_LEAD_SIGN = 0
IIAPI_EPV_MONEY_TRAIL_SIGN = 1
IIAPI_EPV_RET_FATAL = 'fail'
IIAPI_EPV_RET_IGNORE = 'ignore'
IIAPI_EPV_RET_WARN = 'warn'

IIAPI_CPV_ANSIDATE = IIAPI_EPV_ANSIDATE
IIAPI_CPV_BTREE = 'btree'
IIAPI_CPV_CBTREE = 'cbtree'
IIAPI_CPV_CHASH = 'chash'
IIAPI_CPV_CHEAP = 'cheap'
IIAPI_CPV_CISAM = 'cisam'
IIAPI_CPV_DECASDEC = IIAPI_EPV_DECASDEC
IIAPI_CPV_DECASFLOAT = IIAPI_EPV_DECASFLOAT
IIAPI_CPV_DFRMT_ANSI = IIAPI_EPV_DFRMT_ANSI
IIAPI_CPV_DFRMT_DMY = IIAPI_EPV_DFRMT_DMY
IIAPI_CPV_DFRMT_FINNISH = IIAPI_EPV_DFRMT_FINNISH
IIAPI_CPV_DFRMT_GERMAN = IIAPI_EPV_DFRMT_GERMAN
IIAPI_CPV_DFRMT_ISO = IIAPI_EPV_DFRMT_ISO
IIAPI_CPV_DFRMT_ISO4 = IIAPI_EPV_DFRMT_ISO4
IIAPI_CPV_DFRMT_ISO4T = IIAPI_EPV_DFRMT_ISO4T
IIAPI_CPV_DFRMT_ISO4TC = IIAPI_EPV_DFRMT_ISO4TC
IIAPI_CPV_DFRMT_MDY = IIAPI_EPV_DFRMT_MDY
IIAPI_CPV_DFRMT_MLT4 = IIAPI_EPV_DFRMT_MLT4
IIAPI_CPV_DFRMT_MULTI = IIAPI_EPV_DFRMT_MULTI
IIAPI_CPV_DFRMT_US = IIAPI_EPV_DFRMT_US
IIAPI_CPV_DFRMT_YMD = IIAPI_EPV_DFRMT_YMD
IIAPI_CPV_EXPONENTIAL = IIAPI_EPV_EXPONENTIAL
IIAPI_CPV_FLOATDEC = IIAPI_EPV_FLOATDEC
IIAPI_CPV_FLOATDECALIGN = IIAPI_EPV_FLOATDECALIGN
IIAPI_CPV_FLOATF = IIAPI_EPV_FLOATF
IIAPI_CPV_HASH = 'hash'
IIAPI_CPV_HEAP = 'heap'
IIAPI_CPV_INGDATE = IIAPI_EPV_INGDATE
IIAPI_CPV_ISAM = 'isam'
IIAPI_CPV_MONEY_LEAD_SIGN = IIAPI_EPV_MONEY_LEAD_SIGN
IIAPI_CPV_MONEY_TRAIL_SIGN = IIAPI_EPV_MONEY_TRAIL_SIGN
IIAPI_CPV_MONITOR = 1
IIAPI_CPV_RET_FATAL = IIAPI_EPV_RET_FATAL
IIAPI_CPV_RET_IGNORE = IIAPI_EPV_RET_IGNORE
IIAPI_CPV_RET_WARN = IIAPI_EPV_RET_WARN
IIAPI_CPV_SVNORMAL = 0

IIAPI_GE_ERROR = 1
IIAPI_GE_MESSAGE = 3
IIAPI_GE_WARNING = 2
IIAPI_GE_XAERR = 4

IIAPI_GQ_ROW_COUNT = 0x0001
IIAPI_GQ_CURSOR = 0x0002
IIAPI_GQ_PROCEDURE_RET = 0x0004
IIAPI_GQ_PROCEDURE_ID = 0x0008
IIAPI_GQ_REPEAT_QUERY_ID = 0x0010
IIAPI_GQ_TABLE_KEY = 0x0020
IIAPI_GQ_OBJECT_KEY = 0x0040
IIAPI_GQ_ROW_STATUS = 0x0080
IIAPI_GQ_ROW_COUNT_EX = 0x0100

IIAPI_GQF_ALL_UPDATED = 0x0002
IIAPI_GQF_CONTINUE = 0x0020
IIAPI_GQF_END_OF_DATA = 0x0010
IIAPI_GQF_FAIL = 0x0001
IIAPI_GQF_FLUSH_QUERY_ID = 0x0800
IIAPI_GQF_ILLEGAL_XACT_STMT = 0x1000
IIAPI_GQF_INVALID_STATEMENT = 0x0040
IIAPI_GQF_NEW_EFFECTIVE_USER = 0x0400
IIAPI_GQF_NULLS_REMOVED = 0x0004
IIAPI_GQF_OBJECT_KEY = 0x0100
IIAPI_GQF_TABLE_KEY = 0x0200
IIAPI_GQF_TRANSACTION_INACTIVE = 0x0080
IIAPI_GQF_UNKNOWN_REPEAT_QUERY = 0x0008

IIAPI_POS_BEGIN = 0x0001
IIAPI_POS_CURRENT = 0x0003
IIAPI_POS_END = 0x0002

IIAPI_PR_NOECHO = 0x01
IIAPI_PR_TIMEOUT = 0x02
IIAPI_REPLY_TIMEOUT = 0x01

IIAPI_QF_LOCATORS = 0x0001
IIAPI_QF_SCROLL = 0x0002

IIAPI_QT_BASE = 0
IIAPI_QT_QUERY = IIAPI_QT_BASE + 0
IIAPI_QT_ALTER_GROUP = IIAPI_QT_QUERY
IIAPI_QT_ALTER_ROLE = IIAPI_QT_QUERY
IIAPI_QT_ALTER_TABLE = IIAPI_QT_QUERY
IIAPI_QT_COPY_FROM = IIAPI_QT_QUERY
IIAPI_QT_COPY_INTO = IIAPI_QT_QUERY
IIAPI_QT_CREATE_DBEVENT = IIAPI_QT_QUERY
IIAPI_QT_CREATE_GROUP = IIAPI_QT_QUERY
IIAPI_QT_CREATE_INDEX = IIAPI_QT_QUERY
IIAPI_QT_CREATE_INTEGRITY = IIAPI_QT_QUERY
IIAPI_QT_CREATE_PROCEDURE = IIAPI_QT_QUERY
IIAPI_QT_CREATE_ROLE = IIAPI_QT_QUERY
IIAPI_QT_CREATE_RULE = IIAPI_QT_QUERY
IIAPI_QT_CREATE_TABLE = IIAPI_QT_QUERY
IIAPI_QT_CREATE_VIEW = IIAPI_QT_QUERY
IIAPI_QT_CURSOR_DELETE = IIAPI_QT_BASE + 4
IIAPI_QT_CURSOR_UPDATE = IIAPI_QT_BASE + 5
IIAPI_QT_DEF_REPEAT_QUERY = IIAPI_QT_BASE + 6
IIAPI_QT_DELETE = IIAPI_QT_QUERY
IIAPI_QT_DESCRIBE = IIAPI_QT_QUERY
IIAPI_QT_DIRECT_EXEC = IIAPI_QT_QUERY
IIAPI_QT_DROP_DBEVENT = IIAPI_QT_QUERY
IIAPI_QT_DROP_GROUP = IIAPI_QT_QUERY
IIAPI_QT_DROP_INDEX = IIAPI_QT_QUERY
IIAPI_QT_DROP_INTEGRITY = IIAPI_QT_QUERY
IIAPI_QT_DROP_PERMIT = IIAPI_QT_QUERY
IIAPI_QT_DROP_PROCEDURE = IIAPI_QT_QUERY
IIAPI_QT_DROP_ROLE = IIAPI_QT_QUERY
IIAPI_QT_DROP_RULE = IIAPI_QT_QUERY
IIAPI_QT_DROP_TABLE = IIAPI_QT_QUERY
IIAPI_QT_DROP_VIEW = IIAPI_QT_QUERY
IIAPI_QT_EXEC = IIAPI_QT_BASE + 2
IIAPI_QT_EXEC_IMMEDIATE = IIAPI_QT_QUERY
IIAPI_QT_EXEC_PROCEDURE = IIAPI_QT_BASE + 8
IIAPI_QT_EXEC_REPEAT_QUERY = IIAPI_QT_BASE + 7
IIAPI_QT_GET_DBEVENT = IIAPI_QT_QUERY
IIAPI_QT_GRANT = IIAPI_QT_QUERY
IIAPI_QT_INSERT_INTO = IIAPI_QT_QUERY
IIAPI_QT_MODIFY = IIAPI_QT_QUERY
IIAPI_QT_OPEN = IIAPI_QT_BASE + 3
IIAPI_QT_PREPARE = IIAPI_QT_QUERY
IIAPI_QT_RAISE_DBEVENT = IIAPI_QT_QUERY
IIAPI_QT_REGISTER_DBEVENT = IIAPI_QT_QUERY
IIAPI_QT_REMOVE_DBEVENT = IIAPI_QT_QUERY
IIAPI_QT_REVOKE = IIAPI_QT_QUERY
IIAPI_QT_SAVE_TABLE = IIAPI_QT_QUERY
IIAPI_QT_SELECT = IIAPI_QT_QUERY
IIAPI_QT_SELECT_SINGLETON = IIAPI_QT_BASE + 1
IIAPI_QT_SET = IIAPI_QT_QUERY
IIAPI_QT_TOP = IIAPI_QT_EXEC_PROCEDURE
IIAPI_QT_UPDATE = IIAPI_QT_QUERY

IIAPI_ROW_AFTER = 0x0008
IIAPI_ROW_BEFORE = 0x0001
IIAPI_ROW_DELETED = 0x0040
IIAPI_ROW_FIRST = 0x0002
IIAPI_ROW_INSERTED = 0x0010
IIAPI_ROW_LAST = 0x0004
IIAPI_ROW_UPDATED = 0x0020

IIAPI_SCROLL_ABSOLUTE = 0x0008
IIAPI_SCROLL_AFTER = 0x0007
IIAPI_SCROLL_BEFORE = 0x0001
IIAPI_SCROLL_CURRENT = 0x0004
IIAPI_SCROLL_FIRST = 0x0002
IIAPI_SCROLL_LAST = 0x0006
IIAPI_SCROLL_NEXT = 0x0005
IIAPI_SCROLL_PRIOR = 0x0003
IIAPI_SCROLL_RELATIVE = 0x0009

IIAPI_SVR_DEFAULT = 0x0000
IIAPI_SVR_FORMATTED = 0x0010
IIAPI_SVR_MESSAGE = 0x0001
IIAPI_SVR_WARNING = 0x0002

IIAPI_TI_IIXID = 1
IIAPI_TI_XAXID = 2
IIAPI_TRAN_MAXNAME = 64

IIAPI_XA_1PC = 0x40000000
IIAPI_XA_BRANCH_FLAG_1PC = 0x0008
IIAPI_XA_BRANCH_FLAG_2PC = 0x0004
IIAPI_XA_BRANCH_FLAG_FIRST = 0x0001
IIAPI_XA_BRANCH_FLAG_LAST = 0x0002
IIAPI_XA_BRANCH_FLAG_NOOP = 0x0000
IIAPI_XA_FAIL = 0x20000000
IIAPI_XA_HEURCOM = 7
IIAPI_XA_HEURHAZ = 8
IIAPI_XA_HEURMIX = 5
IIAPI_XA_HEURRB = 6
IIAPI_XA_JOIN = 0x00200000
IIAPI_XA_MAXBQUALSIZE = 64
IIAPI_XA_MAXGTRIDSIZE = 64
IIAPI_XA_NOMIGRATE = 9
IIAPI_XA_RBCOMMFAIL = 101
IIAPI_XA_RBDEADLOCK = 102
IIAPI_XA_RBINTEGRITY = 103
IIAPI_XA_RBOTHER = 104
IIAPI_XA_RBPROTO = 105
IIAPI_XA_RBROLLBACK = 100
IIAPI_XA_RBTIMEOUT = 106
IIAPI_XA_RBTRANSIENT = 107
IIAPI_XA_RDONLY = 3
IIAPI_XA_RETRY = 4
IIAPI_XA_XIDDATASIZE = IIAPI_XA_MAXGTRIDSIZE + IIAPI_XA_MAXBQUALSIZE

IIAPI_XAER_ASYNC = -2
IIAPI_XAER_DUPID = -8
IIAPI_XAER_INVAL = -5
IIAPI_XAER_NOTA = -4
IIAPI_XAER_OUTSIDE = -9
IIAPI_XAER_PROTO = -6
IIAPI_XAER_RMERR = -3
IIAPI_XAER_RMFAIL = -7

II_SQLSTATE_LEN = 5
II_SS00000_SUCCESS = '00000'
II_SS01000_WARNING = '01000'
II_SS01001_CURS_OPER_CONFLICT = '01001'
II_SS01002_DISCONNECT_ERROR = '01002'
II_SS01003_NULL_ELIM_IN_SETFUNC = '01003'
II_SS01004_STRING_RIGHT_TRUNC = '01004'
II_SS01005_INSUF_DESCR_AREAS = '01005'
II_SS01006_PRIV_NOT_REVOKED = '01006'
II_SS01007_PRIV_NOT_GRANTED = '01007'
II_SS01008_IMP_ZERO_BIT_PADDING = '01008'
II_SS01009_SEARCH_COND_TOO_LONG = '01009'
II_SS0100A_QRY_EXPR_TOO_LONG = '0100A'
II_SS01500_LDB_TBL_NOT_DROPPED = '01500'
II_SS01501_NO_WHERE_CLAUSE = '01501'
II_SS02000_NO_DATA = '02000'
II_SS07000_DSQL_ERROR = '07000'
II_SS07001_USING_PARM_MISMATCH = '07001'
II_SS07002_USING_TARG_MISMATCH = '07002'
II_SS07003_CAN_EXEC_CURS_SPEC = '07003'
II_SS07004_NEED_USING_FOR_PARMS = '07004'
II_SS07005_STMT_NOT_CURS_SPEC = '07005'
II_SS07006_RESTR_DT_ATTR_ERR = '07006'
II_SS07007_NEED_USING_FOR_RES = '07007'
II_SS07008_INV_DESCR_CNT = '07008'
II_SS07009_INV_DESCR_IDX = '07009'
II_SS07500_CONTEXT_MISMATCH = '07500'
II_SS08000_CONNECTION_EXCEPTION = '08000'
II_SS08001_CANT_GET_CONNECTION = '08001'
II_SS08002_CONNECT_NAME_IN_USE = '08002'
II_SS08003_NO_CONNECTION = '08003'
II_SS08004_CONNECTION_REJECTED = '08004'
II_SS08006_CONNECTION_FAILURE = '08006'
II_SS08007_XACT_RES_UNKNOWN = '08007'
II_SS08500_LDB_UNAVAILABLE = '08500'
II_SS0A000_FEATUR_NOT_SUPPORTED = '0A000'
II_SS0A001_MULT_SERVER_XACTS = '0A001'
II_SS0A500_INVALID_QRY_LANG = '0A500'
II_SS21000_CARD_VIOLATION = '21000'
II_SS22000_DATA_EXCEPTION = '22000'
II_SS22001_STRING_RIGHT_TRUNC = '22001'
II_SS22002_NULLVAL_NO_IND_PARAM = '22002'
II_SS22003_NUM_VAL_OUT_OF_RANGE = '22003'
II_SS22005_ASSIGNMENT_ERROR = '22005'
II_SS22007_INV_DATETIME_FMT = '22007'
II_SS22008_DATETIME_FLD_OVFLOW = '22008'
II_SS22009_INV_TZ_DISPL_VAL = '22009'
II_SS22011_SUBSTRING_ERROR = '22011'
II_SS22012_DIVISION_BY_ZERO = '22012'
II_SS22015_INTERVAL_FLD_OVFLOW = '22015'
II_SS22018_INV_VAL_FOR_CAST = '22018'
II_SS22019_INV_ESCAPE_CHAR = '22019'
II_SS22021_CHAR_NOT_IN_RPRTR = '22021'
II_SS22022_INDICATOR_OVFLOW = '22022'
II_SS22023_INV_PARAM_VAL = '22023'
II_SS22024_UNTERM_C_STRING = '22024'
II_SS22025_INV_ESCAPE_SEQ = '22025'
II_SS22026_STRING_LEN_MISMATCH = '22026'
II_SS22027_TRIM_ERROR = '22027'
II_SS22500_INVALID_DATA_TYPE = '22500'
II_SS23000_CONSTR_VIOLATION = '23000'
II_SS23501_UNIQUE_CONS_VIOLATION = '23501'
II_SS24000_INV_CURS_STATE = '24000'
II_SS25000_INV_XACT_STATE = '25000'
II_SS26000_INV_SQL_STMT_NAME = '26000'
II_SS27000_TRIG_DATA_CHNG_ERR = '27000'
II_SS28000_INV_AUTH_SPEC = '28000'
II_SS2A500_TBL_NOT_FOUND = '2A500'
II_SS2A501_COL_NOT_FOUND = '2A501'
II_SS2A502_DUPL_OBJECT = '2A502'
II_SS2A503_INSUF_PRIV = '2A503'
II_SS2A504_UNKNOWN_CURSOR = '2A504'
II_SS2A505_OBJ_NOT_FOUND = '2A505'
II_SS2A506_INVALID_IDENTIFIER = '2A506'
II_SS2A507_RESERVED_IDENTIFIER = '2A507'
II_SS2B000_DEP_PRIV_EXISTS = '2B000'
II_SS2C000_INV_CH_SET_NAME = '2C000'
II_SS2D000_INV_XACT_TERMINATION = '2D000'
II_SS2E000_INV_CONN_NAME = '2E000'
II_SS33000_INV_SQL_DESCR_NAME = '33000'
II_SS34000_INV_CURS_NAME = '34000'
II_SS35000_INV_COND_NUM = '35000'
II_SS37500_TBL_NOT_FOUND = '37500'
II_SS37501_COL_NOT_FOUND = '37501'
II_SS37502_DUPL_OBJECT = '37502'
II_SS37503_INSUF_PRIV = '37503'
II_SS37504_UNKNOWN_CURSOR = '37504'
II_SS37505_OBJ_NOT_FOUND = '37505'
II_SS37506_INVALID_IDENTIFIER = '37506'
II_SS37507_RESERVED_IDENTIFIER = '37507'
II_SS3C000_AMBIG_CURS_NAME = '3C000'
II_SS3D000_INV_CAT_NAME = '3D000'
II_SS3F000_INV_SCHEMA_NAME = '3F000'
II_SS40000_XACT_ROLLBACK = '40000'
II_SS40001_SERIALIZATION_FAIL = '40001'
II_SS40002_CONSTR_VIOLATION = '40002'
II_SS40003_STMT_COMPL_UNKNOWN = '40003'
II_SS42000_SYN_OR_ACCESSERR = '42000'
II_SS42500_TBL_NOT_FOUND = '42500'
II_SS42501_COL_NOT_FOUND = '42501'
II_SS42502_DUPL_OBJECT = '42502'
II_SS42503_INSUF_PRIV = '42503'
II_SS42504_UNKNOWN_CURSOR = '42504'
II_SS42505_OBJ_NOT_FOUND = '42505'
II_SS42506_INVALID_IDENTIFIER = '42506'
II_SS42507_RESERVED_IDENTIFIER = '42507'
II_SS44000_CHECK_OPTION_ERR = '44000'
II_SS50000_MISC_ING_ERRORS = '50000'
II_SS50001_INVALID_DUP_ROW = '50001'
II_SS50002_LIMIT_EXCEEDED = '50002'
II_SS50003_EXHAUSTED_RESOURCE = '50003'
II_SS50004_SYS_CONFIG_ERROR = '50004'
II_SS50005_GW_ERROR = '50005'
II_SS50006_FATAL_ERROR = '50006'
II_SS50007_INVALID_SQL_STMT_ID = '50007'
II_SS50008_UNSUPPORTED_STMT = '50008'
II_SS50009_ERROR_RAISED_IN_DBPROC = '50009'
II_SS5000A_QUERY_ERROR = '5000A'
II_SS5000B_INTERNAL_ERROR = '5000B'
II_SS5000C_FETCH_ORIENTATION = '5000C'
II_SS5000D_INVALID_CURSOR_NAME = '5000D'
II_SS5000E_DUP_STMT_ID = '5000E'
II_SS5000F_TEXTUAL_INFO = '5000F'
II_SS5000G_DBPROC_MESSAGE = '5000G'
II_SS5000H_UNAVAILABLE_RESOURCE = '5000H'
II_SS5000I_UNEXP_LDB_SCHEMA_CHNG = '5000I'
II_SS5000J_INCONSISTENT_DBMS_CAT = '5000J'
II_SS5000K_SQLSTATE_UNAVAILABLE = '5000K'
II_SS5000L_PROTOCOL_ERROR = '5000L'
II_SS5000M_IPC_ERROR = '5000M'
II_SS5000N_OPERAND_TYPE_MISMATCH = '5000N'
II_SS5000O_INVALID_FUNC_ARG_TYPE = '5000O'
II_SS5000P_TIMEOUT_ON_LOCK_REQUEST = '5000P'
II_SS5000Q_DB_REORG_INVALIDATED_QP = '5000Q'
II_SS5000R_RUN_TIME_LOGICAL_ERROR = '5000R'
II_SSHZ000_RDA = 'HZ000'
